// Copyright 2020 Chaos Mesh Authors.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// See the License for the specific language governing permissions and
// limitations under the License.

package v1alpha1

import (
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
)

var _ = Describe("kernelchaos_webhook", func() {
	Context("Defaulter", func() {
		It("set default namespace selector", func() {
			kernelchaos := &KernelChaos{
				ObjectMeta: metav1.ObjectMeta{Namespace: metav1.NamespaceDefault},
			}
			kernelchaos.Default()
			Expect(kernelchaos.Spec.Selector.Namespaces[0]).To(Equal(metav1.NamespaceDefault))
		})
	})
	Context("ChaosValidator of kernelchaos", func() {
		It("Validate", func() {

			type TestCase struct {
				name    string
				chaos   KernelChaos
				execute func(chaos *KernelChaos) error
				expect  string
			}
			duration := "400s"
			tcs := []TestCase{
				{
					name: "simple ValidateCreate",
					chaos: KernelChaos{
						ObjectMeta: metav1.ObjectMeta{
							Namespace: metav1.NamespaceDefault,
							Name:      "foo1",
						},
					},
					execute: func(chaos *KernelChaos) error {
						return chaos.ValidateCreate()
					},
					expect: "",
				},
				{
					name: "simple ValidateUpdate",
					chaos: KernelChaos{
						ObjectMeta: metav1.ObjectMeta{
							Namespace: metav1.NamespaceDefault,
							Name:      "foo2",
						},
					},
					execute: func(chaos *KernelChaos) error {
						return chaos.ValidateUpdate(chaos)
					},
					expect: "",
				},
				{
					name: "simple ValidateDelete",
					chaos: KernelChaos{
						ObjectMeta: metav1.ObjectMeta{
							Namespace: metav1.NamespaceDefault,
							Name:      "foo3",
						},
					},
					execute: func(chaos *KernelChaos) error {
						return chaos.ValidateDelete()
					},
					expect: "",
				},
				{
					name: "only define the Scheduler",
					chaos: KernelChaos{
						ObjectMeta: metav1.ObjectMeta{
							Namespace: metav1.NamespaceDefault,
							Name:      "foo4",
						},
						Spec: KernelChaosSpec{
							Scheduler: &SchedulerSpec{
								Cron: "@every 10m",
							},
						},
					},
					execute: func(chaos *KernelChaos) error {
						return chaos.ValidateCreate()
					},
					expect: "error",
				},
				{
					name: "only define the Duration",
					chaos: KernelChaos{
						ObjectMeta: metav1.ObjectMeta{
							Namespace: metav1.NamespaceDefault,
							Name:      "foo5",
						},
						Spec: KernelChaosSpec{
							Duration: &duration,
						},
					},
					execute: func(chaos *KernelChaos) error {
						return chaos.ValidateCreate()
					},
					expect: "error",
				},
			}

			for _, tc := range tcs {
				err := tc.execute(&tc.chaos)
				if tc.expect == "error" {
					Expect(err).To(HaveOccurred())
				} else {
					Expect(err).NotTo(HaveOccurred())
				}
			}
		})
	})
})
